@@ -5,7 +5,7 @@
 %%% Created : 23 Aug 2009 by Evgeniy Khramtsov <ekhramtsov@process-one.net>
 %%%
 %%%
-%%% Copyright (C) 2002-2020 ProcessOne, SARL. All Rights Reserved.
+%%% Copyright (C) 2002-2022 ProcessOne, SARL. All Rights Reserved.
 %%%
 %%% Licensed under the Apache License, Version 2.0 (the "License");
 %%% you may not use this file except in compliance with the License.
@@ -37,49 +37,61 @@
 -export([wait_for_allocate/2, active/2]).
 
 -include("stun.hrl").
+-include("stun_logger.hrl").
 
-%%-define(debug, true).
--ifdef(debug).
--define(dbg(Str, Args), error_logger:info_msg(Str, Args)).
--else.
--define(dbg(Str, Args), ok).
--endif.
-
+-define(UDP_READ_PACKETS, 100).
+-define(UDP_ACTIVE, 500).
+-define(UDP_RECBUF, 1024 * 1024). % 1 MiB
 -define(MAX_LIFETIME, 3600000). %% 1 hour
 -define(DEFAULT_LIFETIME, 600000). %% 10 minutes
 -define(PERMISSION_LIFETIME, 300000). %% 5 minutes
 -define(CHANNEL_LIFETIME, 600000). %% 10 minutes
+-define(INITIAL_BLACKLIST, [
+	%% Could be used to circumvent blocking of loopback addresses:
+	{{0, 0, 0, 0}, 8},
+	{{0, 0, 0, 0, 0, 0, 0, 0}, 128},
+	%% RFC 6156, 9.1: "a TURN relay MUST NOT accept Teredo or 6to4 addresses".
+	{{8193, 0, 0, 0, 0, 0, 0, 0}, 32},   % 2001::/32 (Teredo).
+	{{8194, 0, 0, 0, 0, 0, 0, 0}, 16}]). % 2002::/16 (6to4).
 
 -type addr() :: {inet:ip_address(), inet:port_number()}.
 -type subnet() :: {inet:ip4_address(), 0..32} | {inet:ip6_address(), 0..128}.
--type blacklist() :: [subnet()].
+-type accesslist() :: [subnet()].
 
--export_type([blacklist/0]).
+-export_type([accesslist/0]).
 
 -record(state,
-	{sock_mod = gen_udp             :: gen_udp | gen_tcp | fast_tls,
-	 sock                           :: inet:socket() | fast_tls:tls_socket(),
-	 addr = {{0,0,0,0}, 0}          :: addr(),
-	 owner = self()                 :: pid(),
-	 username = <<"">>              :: binary(),
-	 realm = <<"">>                 :: binary(),
-	 key = {<<"">>, <<"">>, <<"">>} :: {binary(), binary(), binary()},
-	 server_name = <<"">>           :: binary(),
-	 peers = #{}                    :: map(),
-	 channels = #{}                 :: map(),
-	 permissions = #{}              :: map(),
-	 max_permissions                :: non_neg_integer() | atom(),
-	 relay_ipv4_ip = {127,0,0,1}    :: inet:ip4_address(),
-	 relay_ipv6_ip                  :: inet:ip6_address(),
-	 min_port = 49152               :: non_neg_integer(),
-	 max_port = 65535               :: non_neg_integer(),
-	 relay_addr                     :: addr(),
-	 relay_sock                     :: inet:socket(),
-	 last_trid                      :: non_neg_integer(),
-	 last_pkt = <<>>                :: binary(),
-	 seq = 1                        :: non_neg_integer(),
-	 life_timer                     :: reference(),
-	 blacklist                      :: blacklist()}).
+	{sock_mod = gen_udp                :: gen_udp | gen_tcp | fast_tls,
+	 sock                              :: inet:socket() | fast_tls:tls_socket() | undefined,
+	 addr = {{0,0,0,0}, 0}             :: addr(),
+	 owner = self()                    :: pid(),
+	 username = <<"">>                 :: binary(),
+	 realm = <<"">>                    :: binary(),
+	 key = {<<"">>, <<"">>, <<"">>}    :: {binary(), binary(), binary()},
+	 server_name = <<"">>              :: binary(),
+	 peers = #{}                       :: map(),
+	 channels = #{}                    :: map(),
+	 permissions = #{}                 :: map(),
+	 max_permissions                   :: non_neg_integer() | atom() | undefined,
+	 relay_ipv4_ip = {127,0,0,1}       :: inet:ip4_address(),
+	 relay_ipv6_ip                     :: inet:ip6_address() | undefined,
+	 min_port = 49152                  :: non_neg_integer(),
+	 max_port = 65535                  :: non_neg_integer(),
+	 relay_addr                        :: addr() | undefined,
+	 relay_sock                        :: inet:socket() | undefined,
+	 last_trid                         :: non_neg_integer() | undefined,
+	 last_pkt = <<>>                   :: binary(),
+	 seq = 1                           :: non_neg_integer(),
+	 life_timer                        :: reference() | undefined,
+	 blacklist = []                    :: accesslist(),
+	 whitelist = []                    :: accesslist(),
+	 hook_fun                          :: function() | undefined,
+	 session_id                        :: binary(),
+	 rcvd_bytes = 0                    :: non_neg_integer(),
+	 rcvd_pkts = 0                     :: non_neg_integer(),
+	 sent_bytes = 0                    :: non_neg_integer(),
+	 sent_pkts = 0                     :: non_neg_integer(),
+	 start_timestamp = get_timestamp() :: integer()}).
 
 %%====================================================================
 %% API
@@ -100,11 +112,17 @@ route(Pid, Msg) ->
 %% gen_fsm callbacks
 %%====================================================================
 init([Opts]) ->
+    process_flag(trap_exit, true),
+    ID = proplists:get_value(session_id, Opts),
     Owner = proplists:get_value(owner, Opts),
     Username = proplists:get_value(username, Opts),
     Realm = proplists:get_value(realm, Opts),
     AddrPort = proplists:get_value(addr, Opts),
-    State = #state{sock_mod = proplists:get_value(sock_mod, Opts),
+    SockMod = proplists:get_value(sock_mod, Opts),
+    HookFun = proplists:get_value(hook_fun, Opts),
+    Blacklist = proplists:get_value(blacklist, Opts) ++ ?INITIAL_BLACKLIST,
+    Whitelist = proplists:get_value(whitelist, Opts),
+    State = #state{sock_mod = SockMod,
 		   sock = proplists:get_value(sock, Opts),
 		   key = proplists:get_value(key, Opts),
 		   relay_ipv4_ip = proplists:get_value(relay_ipv4_ip, Opts),
@@ -112,10 +130,11 @@ init([Opts]) ->
 		   min_port = proplists:get_value(min_port, Opts),
 		   max_port = proplists:get_value(max_port, Opts),
 		   max_permissions = proplists:get_value(max_permissions, Opts),
-		   blacklist = proplists:get_value(blacklist, Opts),
 		   server_name = proplists:get_value(server_name, Opts),
-		   realm = Realm, addr = AddrPort,
-		   username = Username, owner = Owner},
+		   username = Username, realm = Realm, addr = AddrPort,
+		   session_id = ID, owner = Owner, hook_fun = HookFun,
+		   blacklist = Blacklist, whitelist = Whitelist},
+    stun_logger:set_metadata(turn, SockMod, ID, AddrPort, Username),
     MaxAllocs = proplists:get_value(max_allocs, Opts),
     if is_pid(Owner) ->
 	    erlang:monitor(process, Owner);
@@ -125,9 +144,13 @@ init([Opts]) ->
     TRef = erlang:start_timer(?DEFAULT_LIFETIME, self(), stop),
     case turn_sm:add_allocation(AddrPort, Username, Realm, MaxAllocs, self()) of
 	ok ->
-	    {ok, wait_for_allocate, State#state{life_timer = TRef}};
-	{error, Reason} ->
-	    {stop, Reason}
+	    run_hook(turn_session_start, State),
+	    {ok, wait_for_allocate, State#state{life_timer = TRef}}
+	%%
+	%% turn_sm:add_allocation/5 currently doesn't return errors.
+	%%
+	%% {error, Reason} ->
+	%%     {stop, Reason}
     end.
 
 wait_for_allocate(#stun{class = request,
@@ -141,23 +164,30 @@ wait_for_allocate(#stun{class = request,
     IsBlacklisted = blacklisted(State),
     Resp = prepare_response(State, Msg),
     if Msg#stun.'REQUESTED-TRANSPORT' == undefined ->
+	    ?LOG_NOTICE("Rejecting allocation request: no transport requested"),
 	    R = Resp#stun{class = error,
 			  'ERROR-CODE' = stun_codec:error(400)},
 	    {stop, normal, send(State, R)};
        Msg#stun.'REQUESTED-TRANSPORT' == unknown ->
+	    ?LOG_NOTICE("Rejecting allocation request: unsupported transport"),
 	    R = Resp#stun{class = error,
 			  'ERROR-CODE' = stun_codec:error(442)},
 	    {stop, normal, send(State, R)};
        Msg#stun.'DONT-FRAGMENT' == true ->
+	    ?LOG_NOTICE("Rejecting allocation request: dont-fragment not "
+			"supported"),
 	    R = Resp#stun{class = error,
 			  'UNKNOWN-ATTRIBUTES' = [?STUN_ATTR_DONT_FRAGMENT],
 			  'ERROR-CODE' = stun_codec:error(420)},
 	    {stop, normal, send(State, R)};
        Family == inet6, State#state.relay_ipv6_ip == undefined ->
+	    ?LOG_NOTICE("Rejecting allocation request: IPv6 not supported"),
 	    R = Resp#stun{class = error,
 			  'ERROR-CODE' = stun_codec:error(440)},
 	    {stop, normal, send(State, R)};
        IsBlacklisted ->
+	    ?LOG_NOTICE("Rejecting allocation request: Client address is "
+			"blacklisted"),
 	    R = Resp#stun{class = error,
 			  'ERROR-CODE' = stun_codec:error(403)},
 	    {stop, normal, send(State, R)};
@@ -172,9 +202,10 @@ wait_for_allocate(#stun{class = request,
 		    Lifetime = time_left(State#state.life_timer),
 		    AddrPort = stun:unmap_v4_addr(State#state.addr),
 		    RelayAddr = {RelayIP, RelayPort},
-		    ?dbg("created TURN allocation for ~s@~s from ~s: ~s",
-                         [State#state.username, State#state.realm,
-                          addr_to_str(AddrPort), addr_to_str(RelayAddr)]),
+		    stun_logger:add_metadata(
+		      #{stun_relay => stun_logger:encode_addr(RelayAddr)}),
+		    ?LOG_NOTICE("Creating TURN allocation "
+				"(lifetime: ~B seconds)", [Lifetime]),
 		    R = Resp#stun{class = response,
 				  'XOR-RELAYED-ADDRESS' = RelayAddr,
 				  'LIFETIME' = Lifetime,
@@ -184,17 +215,15 @@ wait_for_allocate(#stun{class = request,
 		     NewState#state{relay_sock = RelaySock,
 				    relay_addr = RelayAddr}};
 		Err ->
-		    error_logger:error_msg(
-		      "unable to allocate relay port for ~s@~s: ~s",
-		      [State#state.username, State#state.realm,
-		       format_error(Err)]),
+		    ?LOG_ERROR("Cannot allocate TURN relay: ~s",
+			       [format_error(Err)]),
 		    R = Resp#stun{class = error,
 				  'ERROR-CODE' = stun_codec:error(508)},
 		    {stop, normal, send(State, R)}
 	    end
     end;
 wait_for_allocate(Event, State) ->
-    error_logger:error_msg("unexpected event in wait_for_allocate: ~p", [Event]),
+    ?LOG_ERROR("Unexpected event in 'wait_for_allocate': ~p", [Event]),
     {next_state, wait_for_allocate, State}.
 
 active(#stun{trid = TrID}, #state{last_trid = TrID} = State) ->
@@ -202,6 +231,7 @@ active(#stun{trid = TrID}, #state{last_trid = TrID} = State) ->
     {next_state, active, State};
 active(#stun{class = request,
 	     method = ?STUN_METHOD_ALLOCATE} = Msg, State) ->
+    ?LOG_NOTICE("Rejecting allocation request: Relay already allocated"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'ERROR-CODE' = stun_codec:error(437)},
@@ -209,7 +239,8 @@ active(#stun{class = request,
 active(#stun{class = request,
 	     'REQUESTED-ADDRESS-FAMILY' = ipv4,
 	     method = ?STUN_METHOD_REFRESH} = Msg,
-       #state{relay_addr = {_, _, _, _, _, _, _, _}} = State) ->
+       #state{relay_addr = {{_, _, _, _, _, _, _, _}, _}} = State) ->
+    ?LOG_NOTICE("Rejecting refresh request: IPv4 requested for IPv6 peer"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'ERROR-CODE' = stun_codec:error(443)},
@@ -217,7 +248,8 @@ active(#stun{class = request,
 active(#stun{class = request,
 	     'REQUESTED-ADDRESS-FAMILY' = ipv6,
 	     method = ?STUN_METHOD_REFRESH} = Msg,
-       #state{relay_addr = {_, _, _, _}} = State) ->
+       #state{relay_addr = {{_, _, _, _}, _}} = State) ->
+    ?LOG_NOTICE("Rejecting refresh request: IPv6 requested for IPv4 peer"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'ERROR-CODE' = stun_codec:error(443)},
@@ -227,6 +259,7 @@ active(#stun{class = request,
     Resp = prepare_response(State, Msg),
     case Msg#stun.'LIFETIME' of
 	0 ->
+	    ?LOG_INFO("Client requested closing the TURN session"),
 	    R = Resp#stun{class = response, 'LIFETIME' = 0},
 	    {stop, normal, send(State, R)};
 	LifeTime ->
@@ -236,6 +269,8 @@ active(#stun{class = request,
 		       true ->
 			    lists:min([LifeTime*1000, ?MAX_LIFETIME])
 		    end,
+	    ?LOG_NOTICE("Refreshing TURN allocation (lifetime: ~B seconds)",
+			[MSecs div 1000]),
 	    TRef = erlang:start_timer(MSecs, self(), stop),
 	    R = Resp#stun{class = response,
 			  'LIFETIME' = (MSecs div 1000)},
@@ -251,21 +286,24 @@ active(#stun{class = request,
 	    R = Resp#stun{class = response},
 	    {next_state, active, send(NewState, R)};
 	{error, Code} ->
+	    Err = {_, Txt} = stun_codec:error(Code),
+	    ?LOG_NOTICE("Rejecting permission creation request: ~s", [Txt]),
 	    R = Resp#stun{class = error,
-			  'ERROR-CODE' = stun_codec:error(Code)},
+			  'ERROR-CODE' = Err},
 	    {next_state, active, send(State, R)}
     end;
 active(#stun{class = indication,
 	     method = ?STUN_METHOD_SEND,
 	     'XOR-PEER-ADDRESS' = [{Addr, Port}],
 	     'DATA' = Data}, State) when is_binary(Data) ->
-    case maps:find(Addr, State#state.permissions) of
-	{ok, _} ->
-	    gen_udp:send(State#state.relay_sock, Addr, Port, Data);
-	error ->
-	    ok
-    end,
-    {next_state, active, State};
+    State1 = case maps:find(Addr, State#state.permissions) of
+		 {ok, _} ->
+		     gen_udp:send(State#state.relay_sock, Addr, Port, Data),
+		     count_sent(State, Data);
+		 error ->
+		     State
+	     end,
+    {next_state, active, State1};
 active(#stun{class = request,
 	     'CHANNEL-NUMBER' = Channel,
 	     'XOR-PEER-ADDRESS' = [{Addr, _Port} = Peer],
@@ -275,44 +313,50 @@ active(#stun{class = request,
     case {maps:find(Channel, State#state.channels),
 	  maps:find(Peer, State#state.peers)} of
 	{_, {ok, OldChannel}} when Channel /= OldChannel ->
+	    ?LOG_NOTICE("Rejecting channel binding request: Peer already bound "
+			"to a different channel (~.16B)", [OldChannel]),
 	    R = Resp#stun{class = error,
 			  'ERROR-CODE' = stun_codec:error(400)},
 	    {next_state, active, send(State, R)};
 	{{ok, {OldPeer, _}}, _} when Peer /= OldPeer ->
+	    ?LOG_NOTICE("Rejecting channel binding request: Channel already "
+			"bound to a different peer (~s)",
+			[stun_logger:encode_addr(OldPeer)]),
 	    R = Resp#stun{class = error,
 			  'ERROR-CODE' = stun_codec:error(400)},
 	    {next_state, active, send(State, R)};
 	{FindResult, _} ->
 	    case update_permissions(State, [Addr]) of
 		{ok, NewState0} ->
-		    case FindResult of
-			{ok, {_, OldTRef}} ->
-			    cancel_timer(OldTRef);
-			_ ->
-			    ok
-		    end,
+		    _Op = case FindResult of
+			      {ok, {_, OldTRef}} ->
+				  cancel_timer(OldTRef),
+				  maybe_log(<<"Refreshing">>);
+			      _ ->
+				  maybe_log(<<"Binding">>)
+			  end,
 		    TRef = erlang:start_timer(?CHANNEL_LIFETIME, self(),
 					      {channel_timeout, Channel}),
 		    Peers = maps:put(Peer, Channel, State#state.peers),
 		    Chans = maps:put(Channel, {Peer, TRef},
 				     State#state.channels),
 		    NewState = NewState0#state{peers = Peers, channels = Chans},
-		    ?dbg("bound/refreshed TURN channel ~.16B for user ~s@~s "
-			 "from ~s: ~s <-> ~s",
-			 [Channel, State#state.username, State#state.realm,
-			  addr_to_str(State#state.addr),
-			  addr_to_str(State#state.relay_addr),
-			  addr_to_str(Peer)]),
+		    ?LOG_INFO("~s TURN channel ~.16B for peer ~s",
+			      [_Op, Channel, stun_logger:encode_addr(Peer)]),
 		    R = Resp#stun{class = response},
 		    {next_state, active, send(NewState, R)};
 		{error, Code} ->
+		    Err = {_, Txt} = stun_codec:error(Code),
+		    ?LOG_NOTICE("Rejecting channel binding request: ~s", [Txt]),
 		    R = Resp#stun{class = error,
-				  'ERROR-CODE' = stun_codec:error(Code)},
+				  'ERROR-CODE' = Err},
 		    {next_state, active, send(State, R)}
 	    end
     end;
 active(#stun{class = request,
 	     method = ?STUN_METHOD_CHANNEL_BIND} = Msg, State) ->
+    ?LOG_NOTICE("Rejecting channel binding request: Missing channel number "
+		"and/or peer address"),
     Resp = prepare_response(State, Msg),
     R = Resp#stun{class = error,
 		  'ERROR-CODE' = stun_codec:error(400)},
@@ -322,31 +366,32 @@ active(#turn{channel = Channel, data = Data}, State) ->
 	{ok, {{Addr, Port}, _}} ->
 	    gen_udp:send(State#state.relay_sock,
 			 Addr, Port, Data),
-	    {next_state, active, State};
+	    State1 = count_sent(State, Data),
+	    {next_state, active, State1};
 	error ->
 	    {next_state, active, State}
     end;
 active(Event, State) ->
-    error_logger:error_msg("got unexpected event in active: ~p", [Event]),
+    ?LOG_ERROR("Unexpected event in 'active': ~p", [Event]),
     {next_state, active, State}.
 
 handle_event(stop, _StateName, State) ->
     {stop, normal, State};
 handle_event(Event, StateName, State) ->
-    error_logger:error_msg("got unexpected event in ~s: ~p", [StateName, Event]),
+    ?LOG_ERROR("Unexpected event in '~s': ~p", [StateName, Event]),
     {next_state, StateName, State}.
 
 handle_sync_event(_Event, _From, StateName, State) ->
     {reply, {error, badarg}, StateName, State}.
 
-handle_info({udp, Sock, Addr, Port, Data}, StateName, State) ->
-    inet:setopts(Sock, [{active, once}]),
+handle_info({udp, _Sock, Addr, Port, Data}, StateName, State) ->
     Peer = {Addr, Port},
     case {maps:find(Addr, State#state.permissions),
 	  maps:find(Peer, State#state.peers)} of
 	{{ok, _}, {ok, Channel}} ->
 	    TurnMsg = #turn{channel = Channel, data = Data},
-	    {next_state, StateName, send(State, TurnMsg)};
+	    State1 = count_rcvd(State, Data),
+	    {next_state, StateName, send(State1, TurnMsg)};
 	{{ok, _}, error} ->
 	    Seq = State#state.seq,
 	    Ind = #stun{class = indication,
@@ -354,15 +399,20 @@ handle_info({udp, Sock, Addr, Port, Data}, StateName, State) ->
 			trid = Seq,
 			'XOR-PEER-ADDRESS' = [Peer],
 			'DATA' = Data},
-	    {next_state, StateName, send(State#state{seq = Seq+1}, Ind)};
+	    State1 = count_rcvd(State, Data),
+	    {next_state, StateName, send(State1#state{seq = Seq+1}, Ind)};
 	{error, _} ->
 	    {next_state, StateName, State}
     end;
+handle_info({udp_passive, Sock}, StateName, State) ->
+    inet:setopts(Sock, [{active, ?UDP_ACTIVE}]),
+    {next_state, StateName, State};
 handle_info({timeout, _Tref, stop}, _StateName, State) ->
     {stop, normal, State};
 handle_info({timeout, _Tref, {permission_timeout, Addr}},
 	    StateName, State) ->
-    ?dbg("permission for ~s timed out", [addr_to_str(Addr)]),
+    ?LOG_INFO("TURN permission for ~s timed out",
+	      [stun_logger:encode_addr(Addr)]),
     case maps:find(Addr, State#state.permissions) of
 	{ok, _} ->
 	    Perms = maps:remove(Addr, State#state.permissions),
@@ -372,9 +422,10 @@ handle_info({timeout, _Tref, {permission_timeout, Addr}},
     end;
 handle_info({timeout, _Tref, {channel_timeout, Channel}},
 	    StateName, State) ->
-    ?dbg("channel ~p timed out", [Channel]),
     case maps:find(Channel, State#state.channels) of
 	{ok, {Peer, _}} ->
+	    ?LOG_INFO("TURN channel ~.16B for peer ~s timed out",
+		      [Channel, stun_logger:encode_addr(Peer)]),
 	    Chans = maps:remove(Channel, State#state.channels),
 	    Peers = maps:remove(Peer, State#state.peers),
 	    {next_state, StateName, State#state{channels = Chans,
@@ -385,26 +436,33 @@ handle_info({timeout, _Tref, {channel_timeout, Channel}},
 handle_info({'DOWN', _Ref, _, _, _}, _StateName, State) ->
     {stop, normal, State};
 handle_info(Info, StateName, State) ->
-    error_logger:error_msg("got unexpected info in ~p: ~p", [StateName, Info]),
+    ?LOG_ERROR("Unexpected info in '~s': ~p", [StateName, Info]),
     {next_state, StateName, State}.
 
 terminate(_Reason, _StateName, State) ->
     AddrPort = State#state.addr,
     Username = State#state.username,
     Realm = State#state.realm,
+    RcvdBytes = State#state.rcvd_bytes,
+    RcvdPkts = State#state.rcvd_pkts,
+    SentBytes = State#state.sent_bytes,
+    SentPkts = State#state.sent_pkts,
     case State#state.relay_addr of
 	undefined ->
 	    ok;
 	_RAddrPort ->
-	    ?dbg("deleting TURN allocation for ~s@~s from ~s: ~s",
-                 [Username, Realm, addr_to_str(AddrPort),
-		  addr_to_str(_RAddrPort)])
+	    ?LOG_INFO("Deleting TURN allocation")
     end,
     if is_pid(State#state.owner) ->
 	    stun:stop(State#state.owner);
        true ->
 	    ok
     end,
+    ?LOG_NOTICE("Relayed ~B KiB (in ~B B / ~B packets, out ~B B / ~B packets), "
+		"duration: ~B seconds",
+		[round((RcvdBytes + SentBytes) / 1024), RcvdBytes, RcvdPkts,
+		 SentBytes, SentPkts, get_duration(State, second)]),
+    run_hook(turn_session_stop, State),
     turn_sm:del_allocation(AddrPort, Username, Realm).
 
 code_change(_OldVsn, StateName, State, _Extra) ->
@@ -423,21 +481,18 @@ update_permissions(#state{relay_addr = {IP, _}} = State, Addrs) ->
 	{true, false} ->
 	    Perms = lists:foldl(
 		      fun(Addr, Acc) ->
-			      case maps:find(Addr, Acc) of
-				  {ok, OldTRef} ->
-				      cancel_timer(OldTRef);
-				  error ->
-				      ok
-			      end,
+			      _Op = case maps:find(Addr, Acc) of
+					{ok, OldTRef} ->
+					    cancel_timer(OldTRef),
+					    maybe_log(<<"Refreshing">>);
+					error ->
+					    maybe_log(<<"Creating">>)
+				    end,
 			      TRef = erlang:start_timer(
 				       ?PERMISSION_LIFETIME, self(),
 				       {permission_timeout, Addr}),
-			      ?dbg("created/updated TURN permission for user "
-				   "~s@~s from ~s: ~s <-> ~s",
-				   [State#state.username, State#state.realm,
-				    addr_to_str(State#state.addr),
-				    addr_to_str(State#state.relay_addr),
-				    addr_to_str(Addr)]),
+			      ?LOG_INFO("~s TURN permission for ~s",
+					[_Op, stun_logger:encode_addr(Addr)]),
 			      maps:put(Addr, TRef, Acc)
 		      end, State#state.permissions, Addrs),
 	    {ok, State#state{permissions = Perms}};
@@ -455,12 +510,15 @@ send(State, Pkt) when is_binary(Pkt) ->
 	    gen_udp:send(Sock, Addr, Port, Pkt);
        true ->
 	    case SockMod:send(Sock, Pkt) of
-		ok -> ok;
-		_  -> exit(normal)
+		ok ->
+		    ok;
+		_  ->
+		    ?LOG_INFO("Cannot respond to client: Connection closed"),
+		    exit(normal)
 	    end
     end;
 send(State, Msg) ->
-    ?dbg("send:~n~s", [stun_codec:pp(Msg)]),
+    ?LOG_DEBUG(#{verbatim => {"Sending:~n~s", [stun_codec:pp(Msg)]}}),
     Key = State#state.key,
     case Msg of
 	#stun{class = indication} ->
@@ -489,7 +547,12 @@ allocate_addr(Family, Addr, {Min, Max}) ->
 allocate_addr(_Family, _Addr, _Min, _Max, _Next, 0) ->
     {error, eaddrinuse};
 allocate_addr(Family, Addr, Min, Max, Next, Count) ->
-    case gen_udp:open(Next, [binary, Family, {ip, Addr}, {active, once}]) of
+    case gen_udp:open(Next, [binary,
+			     Family,
+			     {ip, Addr},
+			     {active, ?UDP_ACTIVE},
+			     {recbuf, ?UDP_RECBUF},
+			     {read_packets, ?UDP_READ_PACKETS}]) of
 	{ok, Sock} ->
 	    case inet:sockname(Sock) of
 		{ok, {_, Port}} ->
@@ -522,14 +585,22 @@ family_matches(_Addr1, _Addr2) ->
 blacklisted(#state{addr = {IP, _Port}} = State) ->
     blacklisted(State, [IP]).
 
-blacklisted(#state{blacklist = Blacklist}, IPs) ->
+blacklisted(#state{blacklist = Blacklist, whitelist = Whitelist}, IPs) ->
     lists:any(
       fun(IP) ->
 	      lists:any(
 		fun({Net, Mask}) ->
 			match_subnet(IP, Net, Mask)
 		end, Blacklist)
-      end, IPs).
+      end, IPs)
+          andalso not
+	      lists:any(
+		fun(IP) ->
+			lists:any(
+			  fun({Net, Mask}) ->
+				  match_subnet(IP, Net, Mask)
+			  end, Whitelist)
+		end, IPs).
 
 match_subnet({_, _, _, _} = IP,
 	     {_, _, _, _} = Net, Mask) ->
@@ -572,15 +643,6 @@ format_error({error, Reason}) ->
 	    Res
     end.
 
--ifdef(debug).
-addr_to_str({{_, _, _, _, _, _, _, _} = Addr, Port}) ->
-    [$[, inet_parse:ntoa(Addr), $], $:, integer_to_list(Port)];
-addr_to_str({{_, _, _, _} = Addr, Port}) ->
-    [inet_parse:ntoa(Addr), $:, integer_to_list(Port)];
-addr_to_str(Addr) ->
-    inet_parse:ntoa(Addr).
--endif.
-
 cancel_timer(undefined) ->
     ok;
 cancel_timer(TRef) ->
@@ -596,8 +658,70 @@ cancel_timer(TRef) ->
             ok
     end.
 
+get_timestamp() ->
+    erlang:monotonic_time().
+
+get_duration(State, Unit) ->
+    erlang:convert_time_unit(get_duration(State), native, Unit).
+
+get_duration(#state{start_timestamp = Start}) ->
+    get_timestamp() - Start.
+
 prepare_response(State, Msg) ->
     #stun{method = Msg#stun.method,
 	  magic = Msg#stun.magic,
 	  trid = Msg#stun.trid,
 	  'SOFTWARE' = State#state.server_name}.
+
+count_sent(#state{sent_bytes = SentSize,
+		  sent_pkts = SentPkts} = State, Data) ->
+    State#state{sent_bytes = SentSize + byte_size(Data),
+		sent_pkts = SentPkts + 1}.
+
+count_rcvd(#state{rcvd_bytes = RcvdSize,
+		  rcvd_pkts = RcvdPkts} = State, Data) ->
+    State#state{rcvd_bytes = RcvdSize + byte_size(Data),
+		rcvd_pkts = RcvdPkts + 1}.
+
+run_hook(HookName, #state{session_id = ID,
+			  username = User,
+			  realm = Realm,
+			  addr = Client,
+			  sock_mod = SockMod,
+			  hook_fun = HookFun} = State)
+  when is_function(HookFun) ->
+    Info0 = #{id => ID,
+	      user => User,
+	      realm => Realm,
+	      client => Client,
+	      transport => stun_logger:encode_transport(SockMod)},
+    Info = case {HookName, State} of
+	       {turn_session_start, _State} ->
+		   Info0;
+	       {turn_session_stop, #state{sent_bytes = SentBytes,
+					  sent_pkts = SentPkts,
+					  rcvd_bytes = RcvdBytes,
+					  rcvd_pkts = RcvdPkts}} ->
+		   Info0#{sent_bytes => SentBytes,
+			  sent_pkts => SentPkts,
+			  rcvd_bytes => RcvdBytes,
+			  rcvd_pkts => RcvdPkts,
+			  duration => get_duration(State)}
+	   end,
+    ?LOG_DEBUG("Running '~s' hook", [HookName]),
+    try HookFun(HookName, Info)
+    catch _:Err -> ?LOG_ERROR("Hook '~s' failed: ~p", [HookName, Err])
+    end;
+run_hook(HookName, _State) ->
+    ?LOG_DEBUG("No callback function specified for '~s' hook", [HookName]),
+    ok.
+
+-ifdef(USE_OLD_LOGGER).
+-ifdef(debug).
+maybe_log(Term) -> Term.
+-else.
+maybe_log(_Term) -> ok.
+-endif.
+-else.
+maybe_log(Term) -> Term.
+-endif.
